/*
 * Author:      William Chia-Wei Cheng (william@cs.ucla.edu)
 *
 * Copyright (C) 1990-2000, William Chia-Wei Cheng.
 *
 * Permission limited to the use, copy, display, distribute without
 * charging for a fee, and produce derivative works of "tgif" and
 * its documentation for not-for-profit purpose is hereby granted by
 * the Author, provided that the above copyright notice appears in
 * all copies made of "tgif" and that both the copyright notice
 * and this permission notice appear in supporting documentation,
 * and that the name of the Author not be used in advertising or
 * publicity pertaining to distribution of the software without
 * specific, written prior permission.  The Author makes no
 * representations about the suitability of this software for any
 * purpose.  It is provided "as is" without express or implied
 * warranty.  All other rights (including, but not limited to, the
 * right to sell "tgif", the right to sell or distribute derivative
 * works of "tgif", the right to distribute "tgif" for a fee, and
 * the right to include "tgif" or derivative works of "tgif" in a
 * for-sale product or service) are reserved by the Author.
 *
 * THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS,
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, INDIRECT
 * OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * @(#)$Header: /mm/src/tgif/v4/RCS/inmethod.c,v 4.8 2000/01/15 22:23:49 william Exp $
 */

#define _INCLUDE_FROM_INMETHOD_C_

#include "tgifdefs.h"

#ifndef _NO_XCIN
#include "cli_xcin.e"
#endif /* ~_NO_XCIN */
#ifndef _NO_CHINPUT
#include "chinput.e"
#endif /* ~_NO_CHINPUT */
#ifndef _NO_KINPUT
#include "convkinput.e"
#endif /* ~_NO_KINPUT */
#ifndef _NO_XIM
#include "convxim.e"
#endif /* ~_NO_XIM */
#include "inmethod.e"
#include "util.e"
#include "setup.e"

int gnInputMethod=TGIM_NONE; /* one of TGIM_* */
int gnOverTheSpot=FALSE;

int cmdLineHasInputMethod=FALSE;
char cmdLineInputMethod[MAXSTRING];

static int gnInputMethodIndex=INVALID;

typedef void (CleanUpFunc)ARGS_DECL((Display *dpy, Window win));
typedef int (InitFunc)ARGS_DECL((Display *dpy, Window win));
typedef int (FocusInFunc)ARGS_DECL((Display *dpy, Window win));
typedef int (FocusOutFunc)ARGS_DECL((Display *dpy, Window win));
typedef int (SendKeyFunc)ARGS_DECL((Display *dpy, Window win,
                                    XKeyEvent *key_ev, char *buf));
typedef int (ConvPropertyFunc)ARGS_DECL((Display *dpy, Window win,
                                         XPropertyEvent *prop_ev,
                                         char **ppsz_buf));
typedef int (ExpectCMFunc)ARGS_DECL((Display *dpy, Window win));
typedef int (HandleCMFunc)ARGS_DECL((Display *dpy, Window win,
                                     XClientMessageEvent *cm_ev,
                                     XKeyEvent *key_ev, char *buf));
typedef int (ActivateOnCntrlSpaceFunc)ARGS_DECL((Display *dpy, Window win));
typedef int (HandleCntrlSpaceFunc)ARGS_DECL((Display *dpy, Window win));
typedef int (DeactivateOnCreateTextFunc)ARGS_DECL((Display *dpy, Window win));
typedef int (HandleCreateTextFunc)ARGS_DECL((Display *dpy, Window win));
typedef int (TellCursorPosFunc)ARGS_DECL((Display *dpy, Window win,
                                          int cur_x, int cur_y));
typedef int (ExpectNextEventFunc)ARGS_DECL((Display *dpy, Window win));
typedef int (HandleNextEventFunc)ARGS_DECL((Display *dpy, Window win,
                                            XEvent *ev));
typedef int (HandleNewCurTextFunc)ARGS_DECL((Display *dpy, Window win));
typedef int (ExpectLookupStringFunc)ARGS_DECL((Display *dpy, Window win));
typedef int (HandleLookupStringFunc)ARGS_DECL((Display *dpy, Window win,
                                               XKeyEvent *key_ev, char *psz_buf,
                                               int buf_size, KeySym *p_key_sym,
                                               int *pn_has_ch));

typedef struct IMInfoRec {
   char *pszName;
   int nInputMethod;
   CleanUpFunc *pCleanUpFunc;
   InitFunc *pInitFunc;
   FocusInFunc *pFocusInFunc;
   FocusOutFunc *pFocusOutFunc;
   SendKeyFunc *pSendKeyFunc;
   ConvPropertyFunc *pConvPropertyFunc;
   ExpectCMFunc *pExpectCMFunc;
   HandleCMFunc *pHandleCMFunc;
   ActivateOnCntrlSpaceFunc *pActivateOnCntrlSpaceFunc;
   HandleCntrlSpaceFunc *pHandleCntrlSpaceFunc;
   DeactivateOnCreateTextFunc *pDeactivateOnCreateTextFunc;
   HandleCreateTextFunc *pHandleCreateTextFunc;
   TellCursorPosFunc *pTellCursorPosFunc;
   ExpectNextEventFunc *pExpectNextEventFunc;
   HandleNextEventFunc *pHandleNextEventFunc;
   HandleNewCurTextFunc *pHandleNewCurTextFunc;
   ExpectLookupStringFunc *pExpectLookupStringFunc;
   HandleLookupStringFunc *pHandleLookupStringFunc;
} * IMInfoPtr;

/* --------------------- xcin Routines --------------------- */

#ifndef _NO_XCIN
static
int XcinInit(dpy, win)
   Display *dpy;
   Window win;
{
   return TRUE;
}

static
int XcinFocusIn(dpy, win)
   Display *dpy;
   Window win;
{
   send_FocusIn(dpy, win);
   return TRUE;
}

static
int XcinFocusOut(dpy, win)
   Display *dpy;
   Window win;
{
   send_FocusOut(dpy, win);
   return TRUE;
}

static
int XcinSendKey(dpy, win, key_ev, buf)
   Display *dpy;
   Window win;
   XKeyEvent *key_ev;
   char *buf;
{
   return send_key(dpy, win, key_ev, buf);
}

static
int XcinExpectCM(dpy, win)
   Display *dpy;
   Window win;
{
   return FALSE;
}
#endif /* ~_NO_XCIN */

/* --------------------- chinput Routines --------------------- */

#ifndef _NO_CHINPUT
static
int ChinputInit(dpy, win)
   Display *dpy;
   Window win;
{
   return (HZclientInit(dpy)==0);
}

static
int ChinputFocusIn(dpy, win)
   Display *dpy;
   Window win;
{
   return TRUE;
}

static
int ChinputFocusOut(dpy, win)
   Display *dpy;
   Window win;
{
   return TRUE;
}

static
int ChinputSendKey(dpy, win, key_ev, buf)
   Display *dpy;
   Window win;
   XKeyEvent *key_ev;
   char *buf;
{
   HZsendKey(dpy, win, key_ev);
   *buf = '\0';
   return TRUE;
}

static
int ChinputExpectCM(dpy, win)
   Display *dpy;
   Window win;
{
   return TRUE;
}

static
int ChinputHandleCM(dpy, win, cm_ev, key_ev, buf)
   Display *dpy;
   Window win;
   XClientMessageEvent *cm_ev;
   XKeyEvent *key_ev;
   char *buf;
{
   char s1[80];
   int num=0, rc=0;

   rc = HZprocInput(cm_ev, &num, &key_ev->keycode, &key_ev->state, s1);

   if (rc == HZSERVER_OUTPUT_BAD) {
      return FALSE;
   } else if (rc == HZSERVER_OUTPUT_NORMAL_KEY) {
      buf[0] = key_ev->keycode;
      buf[1] = '\0';
   } else if (rc == HZSERVER_OUTPUT_FUNCTION_KEY) {
      buf[0] = '\0';
   } else if (rc == HZSERVER_OUTPUT_HANZI) {
      strncpy(buf, s1, num);
      buf[num] = '\0';
   }
   return TRUE;
}
#endif /* ~_NO_CHINPUT */

/* --------------------- kinput Routines --------------------- */

#ifndef _NO_KINPUT
static
void KinputCleanUp(dpy, win)
   Display *dpy;
   Window win;
{
   imProtocol = IM_NONE;
}

static
int KinputInit(dpy, win)
   Display *dpy;
   Window win;
{
   char *c_ptr=NULL;

   copyAndPasteJIS = FALSE;
   if ((c_ptr=XGetDefault(mainDisplay, TOOL_NAME, "CopyAndPasteJIS")) != NULL &&
         UtilStrICmp(c_ptr, "true") == 0) {
      copyAndPasteJIS = TRUE;
   }
   gnOverTheSpot = FALSE;
   if ((c_ptr=XGetDefault(mainDisplay, TOOL_NAME, "PreeditType")) != NULL &&
         UtilStrICmp(c_ptr, "overthespot") == 0) {
      gnOverTheSpot = TRUE;
   }
   KinputSetConvOverSpot(gnOverTheSpot);

   if ((c_ptr=XGetDefault(mainDisplay, TOOL_NAME, "ConvSelection")) != NULL) {
      UtilTrimBlanks(c_ptr);
      strncpy(kinputConvSelName, c_ptr, MAXSTRING-1);
      kinputConvSelName[MAXSTRING-1] = '\0';
   }

   imProtocol = IM_KINPUT;
   return TRUE;
}

static
int KinputFocusIn(dpy, win)
   Display *dpy;
   Window win;
{
   return TRUE;
}

static
int KinputFocusOut(dpy, win)
   Display *dpy;
   Window win;
{
   return TRUE;
}

static
int KinputSendKey(dpy, win, key_ev, buf)
   Display *dpy;
   Window win;
   XKeyEvent *key_ev;
   char *buf;
{
   return FALSE;
}

static
int KinputConvProperty(dpy, win, prop_ev, ppsz_buf)
   Display *dpy;
   Window win;
   XPropertyEvent *prop_ev;
   char **ppsz_buf;
{
   char *c_ptr=KinputCheckConvProperty(dpy, win, prop_ev);

   if (c_ptr != NULL && ppsz_buf != NULL) {
      *ppsz_buf = c_ptr;
      return TRUE;
   }
   return FALSE;
}

static
int KinputExpectCM(dpy, win)
   Display *dpy;
   Window win;
{
   return (win==drawWindow);
}

static
int KinputHandleCM(dpy, win, cm_ev, key_ev, buf)
   Display *dpy;
   Window win;
   XClientMessageEvent *cm_ev;
   XKeyEvent *key_ev;
   char *buf;
{
   KinputCheckClientMessage(dpy, win, cm_ev);
   return FALSE;
}

static
int KinputActiveOnCntrlSpace(dpy, win)
   Display *dpy;
   Window win;
{
   return (win==drawWindow);
}

static
int KinputHandleCntrlSpace(dpy, win)
   Display *dpy;
   Window win;
{
   KinputBeginConversion(dpy, win);
   return TRUE;
}

static
int KinputDeactiveOnCreateText(dpy, win)
   Display *dpy;
   Window win;
{
   return TRUE;
}

static
int KinputHandleCreateText(dpy, win)
   Display *dpy;
   Window win;
{
   KinputEndConversion(dpy, win);
   return TRUE;
}

static
int KinputTellCursorPos(dpy, win, cur_x, cur_y)
   Display *dpy;
   Window win;
   int cur_x, cur_y;
{
   KinputTellCursorPosition(dpy, win, cur_x, cur_y);
   return TRUE;
}
#endif /* ~_NO_KINPUT */

/* --------------------- xim Routines --------------------- */

#ifndef _NO_XIM
static
void XimCleanUp(dpy, win)
   Display *dpy;
   Window win;
{
   XIMClose();
}

static
int XimInit(dpy, win)
   Display *dpy;
   Window win;
{
   char *c_ptr=NULL;
   gnOverTheSpot = FALSE;
   if ((c_ptr=XGetDefault(mainDisplay, TOOL_NAME, "PreeditType")) != NULL &&
         UtilStrICmp(c_ptr, "overthespot") == 0) {
      gnOverTheSpot = TRUE;
   }
   XIMSetConvOverSpot(gnOverTheSpot);
   return TRUE;
}

static
int XimDeactiveOnCreateText(dpy, win)
   Display *dpy;
   Window win;
{
   return TRUE;
}

static
int XimHandleCreateText(dpy, win)
   Display *dpy;
   Window win;
{
   XIMUnsetICFocus();
   return TRUE;
}

static
int XimExpectNextEvent(dpy, win)
   Display *dpy;
   Window win;
{
   return (win==drawWindow);
}

static
int XimHandleNextEvent(dpy, win, ev)
   Display *dpy;
   Window win;
   XEvent *ev;
{
   XIMNextEvent(dpy, win, ev);
   return TRUE;
}

static
int XimHandleNewCurText(dpy, win)
   Display *dpy;
   Window win;
{
   XIMSetICFocus(dpy, win);
   return TRUE;
}

static
int XimExpectLookupString(dpy, win)
   Display *dpy;
   Window win;
{
   return (win==drawWindow);
}

static XIMStatus c_stat;

static
int XimHandleLookupString(dpy, win, key_ev, psz_buf, buf_size, p_key_sym,
      pn_has_ch)
   Display *dpy;
   Window win;
   XKeyEvent *key_ev;
   char *psz_buf;
   int buf_size;
   KeySym *p_key_sym;
   int *pn_has_ch;
{
   *pn_has_ch = XIMLookupString(key_ev, psz_buf, buf_size, p_key_sym, &c_stat);
   return TRUE;
}

static
int XimTellCursorPos(dpy, win, cur_x, cur_y)
   Display *dpy;
   Window win;
   int cur_x, cur_y;
{
   XIMTellCursorPosition(dpy, win, cur_x, cur_y);
   return TRUE;
}
#endif /* ~_NO_XIM */

/* --------------------- Global Routines --------------------- */

static struct IMInfoRec gstIMInfo[] = {
#ifndef _NO_XCIN
   { "xcin", TGIM_XCIN,
     NULL, XcinInit, XcinFocusIn, XcinFocusOut,
     XcinSendKey, NULL, XcinExpectCM, NULL,
     NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL
   },
#endif /* ~_NO_XCIN */
#ifndef _NO_CHINPUT
   { "chinput", TGIM_CHINPUT,
     NULL, ChinputInit, ChinputFocusIn, ChinputFocusOut,
     ChinputSendKey, NULL, ChinputExpectCM, ChinputHandleCM,
     NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL
   },
#endif /* ~_NO_CHINPUT */
#ifndef _NO_KINPUT
   { "kinput2", TGIM_KINPUT,
     KinputCleanUp, KinputInit, KinputFocusIn, KinputFocusOut,
     KinputSendKey, KinputConvProperty, KinputExpectCM, KinputHandleCM,
     KinputActiveOnCntrlSpace, KinputHandleCntrlSpace,
     KinputDeactiveOnCreateText, KinputHandleCreateText,
     KinputTellCursorPos, NULL, NULL, NULL, NULL, NULL
   },
#endif /* ~_NO_KINPUT */
#ifndef _NO_XIM
   { "xim", TGIM_XIM,
     XimCleanUp, XimInit, NULL, NULL,
     NULL, NULL, NULL, NULL, NULL, NULL,
     XimDeactiveOnCreateText, XimHandleCreateText,
     XimTellCursorPos, XimExpectNextEvent, XimHandleNextEvent,
     XimHandleNewCurText, XimExpectLookupString, XimHandleLookupString
   },
#endif /* ~_NO_XIM */
   { NULL, TGIM_NONE, NULL, NULL, NULL, NULL,
     NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
     NULL, NULL, NULL, NULL, NULL
   }
};

void CleanUpInputMethods()
{
   if (gnInputMethodIndex != INVALID &&
         gstIMInfo[gnInputMethodIndex].pCleanUpFunc != NULL) {
      (gstIMInfo[gnInputMethodIndex].pCleanUpFunc)(mainDisplay, mainWindow);
   }
   gnInputMethodIndex = INVALID;
   cmdLineHasInputMethod = FALSE;
   *cmdLineInputMethod = '\0';
}

int InitInputMethods()
{
   char *psz=NULL;

   gnInputMethod = gnInputMethodIndex = INVALID;
   if (cmdLineHasInputMethod) {
      UtilTrimBlanks(cmdLineInputMethod);
      psz = cmdLineInputMethod;
   } else {
      psz = XGetDefault(mainDisplay, TOOL_NAME, "DoubleByteInputMethod");
   }
   if (psz != NULL) {
      struct IMInfoRec *pIMInfo=gstIMInfo;
      int i=0;

      for ( ; pIMInfo->pszName != NULL; pIMInfo++, i++) {
         if (UtilStrICmp(psz, pIMInfo->pszName) == 0) {
            gnInputMethod = pIMInfo->nInputMethod;
            gnInputMethodIndex = i;
            if (gstIMInfo[gnInputMethodIndex].pInitFunc != NULL &&
                  !((gstIMInfo[gnInputMethodIndex].pInitFunc)(mainDisplay,
                  mainWindow))) {
               /* may not be a problem */
            }
            break;
         }
      }
   }
   return TRUE;
}

int tgIMFocusIn(dpy, win)
   Display *dpy;
   Window win;
{
   if (gnInputMethodIndex == INVALID ||
         gstIMInfo[gnInputMethodIndex].pFocusInFunc == NULL) {
      return FALSE;
   }
   return ((gstIMInfo[gnInputMethodIndex].pFocusInFunc)(dpy, win));
}

int tgIMFocusOut(dpy, win)
   Display *dpy;
   Window win;
{
   if (gnInputMethodIndex == INVALID ||
         gstIMInfo[gnInputMethodIndex].pFocusOutFunc == NULL) {
      return FALSE;
   }
   return ((gstIMInfo[gnInputMethodIndex].pFocusOutFunc)(dpy, win));
}

int tgIMTranslateKeyEvent(dpy, win, key_ev, buf)
   Display *dpy;
   Window win;
   XKeyEvent *key_ev;
   char *buf;
{
   if (gnInputMethodIndex == INVALID ||
         gstIMInfo[gnInputMethodIndex].pSendKeyFunc == NULL) {
      return FALSE;
   }
   return ((gstIMInfo[gnInputMethodIndex].pSendKeyFunc)(dpy, win, key_ev, buf));
}

int tgIMConvertProperty(dpy, win, prop_ev, ppsz_buf)
   Display *dpy;
   Window win;
   XPropertyEvent *prop_ev;
   char **ppsz_buf;
{
   if (gnInputMethodIndex == INVALID ||
         gstIMInfo[gnInputMethodIndex].pConvPropertyFunc == NULL) {
      return FALSE;
   }
   return ((gstIMInfo[gnInputMethodIndex].pConvPropertyFunc)(dpy, win, prop_ev,
         ppsz_buf));
}

int tgIMExpectClientMessage(dpy, win)
   Display *dpy;
   Window win;
{
   if (gnInputMethodIndex == INVALID ||
         gstIMInfo[gnInputMethodIndex].pExpectCMFunc == NULL) {
      return FALSE;
   }
   return ((gstIMInfo[gnInputMethodIndex].pExpectCMFunc)(dpy, win));
}

int tgIMHandleClientMessage(dpy, win, cm_ev, key_ev, buf)
   Display *dpy;
   Window win;
   XClientMessageEvent *cm_ev;
   XKeyEvent *key_ev;
   char *buf;
{
   if (gnInputMethodIndex == INVALID ||
         gstIMInfo[gnInputMethodIndex].pHandleCMFunc == NULL) {
      return FALSE;
   }
   return ((gstIMInfo[gnInputMethodIndex].pHandleCMFunc)(dpy, win, cm_ev,
         key_ev, buf));
}

int tgIMActiveOnCntrlSpace(dpy, win)
   Display *dpy;
   Window win;
{
   if (gnInputMethodIndex == INVALID ||
         gstIMInfo[gnInputMethodIndex].pActivateOnCntrlSpaceFunc == NULL) {
      return FALSE;
   }
   return ((gstIMInfo[gnInputMethodIndex].pActivateOnCntrlSpaceFunc)(dpy, win));
}

int tgIMHandleCntrlSpace(dpy, win)
   Display *dpy;
   Window win;
{
   if (gnInputMethodIndex == INVALID ||
         gstIMInfo[gnInputMethodIndex].pHandleCntrlSpaceFunc == NULL) {
      return FALSE;
   }
   return ((gstIMInfo[gnInputMethodIndex].pHandleCntrlSpaceFunc)(dpy, win));
}

int tgIMDeactiveOnCreateText(dpy, win)
   Display *dpy;
   Window win;
{
   if (gnInputMethodIndex == INVALID ||
         gstIMInfo[gnInputMethodIndex].pDeactivateOnCreateTextFunc == NULL) {
      return FALSE;
   }
   return ((gstIMInfo[gnInputMethodIndex].pDeactivateOnCreateTextFunc)(dpy,
         win));
}

int tgIMHandleCreateText(dpy, win)
   Display *dpy;
   Window win;
{
   if (gnInputMethodIndex == INVALID ||
         gstIMInfo[gnInputMethodIndex].pHandleCreateTextFunc == NULL) {
      return FALSE;
   }
   return ((gstIMInfo[gnInputMethodIndex].pHandleCreateTextFunc)(dpy, win));
}

int tgIMTellCursorPosition(dpy, win, cur_x, cur_y)
   Display *dpy;
   Window win;
   int cur_x, cur_y;
{
   if (gnInputMethodIndex == INVALID ||
         gstIMInfo[gnInputMethodIndex].pTellCursorPosFunc == NULL) {
      return FALSE;
   }
   return ((gstIMInfo[gnInputMethodIndex].pTellCursorPosFunc)(dpy, win,
         cur_x, cur_y));
}

int tgIMExpectNextEvent(dpy, win)
   Display *dpy;
   Window win;
{
   if (gnInputMethodIndex == INVALID ||
         gstIMInfo[gnInputMethodIndex].pExpectNextEventFunc == NULL) {
      return FALSE;
   }
   return ((gstIMInfo[gnInputMethodIndex].pExpectNextEventFunc)(dpy, win));
}

int tgIMHandleNextEvent(dpy, win, ev)
   Display *dpy;
   Window win;
   XEvent *ev;
{
   if (gnInputMethodIndex == INVALID ||
         gstIMInfo[gnInputMethodIndex].pHandleNextEventFunc == NULL) {
      return FALSE;
   }
   return ((gstIMInfo[gnInputMethodIndex].pHandleNextEventFunc)(dpy, win, ev));
}

int tgIMHandleNewCurText(dpy, win)
   Display *dpy;
   Window win;
{
   if (gnInputMethodIndex == INVALID ||
         gstIMInfo[gnInputMethodIndex].pHandleNewCurTextFunc == NULL) {
      return FALSE;
   }
   return ((gstIMInfo[gnInputMethodIndex].pHandleNewCurTextFunc)(dpy, win));
}

int tgIMExpectLookupString(dpy, win)
   Display *dpy;
   Window win;
{
   if (gnInputMethodIndex == INVALID ||
         gstIMInfo[gnInputMethodIndex].pExpectLookupStringFunc == NULL) {
      return FALSE;
   }
   return ((gstIMInfo[gnInputMethodIndex].pExpectLookupStringFunc)(dpy, win));
}

int tgIMHandleLookupString(dpy, win, key_ev, psz_buf, buf_size, p_key_sym,
      pn_has_ch)
   Display *dpy;
   Window win;
   XKeyEvent *key_ev;
   char *psz_buf;
   int buf_size;
   KeySym *p_key_sym;
   int *pn_has_ch;
{
   if (gnInputMethodIndex == INVALID ||
         gstIMInfo[gnInputMethodIndex].pHandleLookupStringFunc == NULL) {
      return FALSE;
   }
   return ((gstIMInfo[gnInputMethodIndex].pHandleLookupStringFunc)(dpy, win,
         key_ev, psz_buf, buf_size, p_key_sym, pn_has_ch));
}

