/*
 * Author:      Daniel Ratton Figueiredo <ratton@land.ufrj.br>
 * Author:      William Chia-Wei Cheng (william@cs.ucla.edu)
 *
 * Copyright (C) 1990-2000, William Chia-Wei Cheng.
 *
 * Permission limited to the use, copy, display, distribute without
 * charging for a fee, and produce derivative works of "tgif" and
 * its documentation for not-for-profit purpose is hereby granted by
 * the Author, provided that the above copyright notice appears in
 * all copies made of "tgif" and that both the copyright notice
 * and this permission notice appear in supporting documentation,
 * and that the name of the Author not be used in advertising or
 * publicity pertaining to distribution of the software without
 * specific, written prior permission.  The Author makes no
 * representations about the suitability of this software for any
 * purpose.  It is provided "as is" without express or implied
 * warranty.  All other rights (including, but not limited to, the
 * right to sell "tgif", the right to sell or distribute derivative
 * works of "tgif", the right to distribute "tgif" for a fee, and
 * the right to include "tgif" or derivative works of "tgif" in a
 * for-sale product or service) are reserved by the Author.
 *
 * THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS,
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, INDIRECT
 * OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * @(#)$Header: /mm/src/tgif/v4/RCS/tangram2.c,v 4.17 2000/04/18 00:27:53 william Exp $
 */

#define _INCLUDE_FROM_TANGRAM2_C_

#include "tgifdefs.h"

#include "attr.e"
#include "auxtext.e"
#include "box.e"
#include "cmd.e"
#include "dialog.e"
#include "drawing.e"
#include "exec.e"
#include "menu.e"
#include "msg.e"
#include "miniline.e"
#include "move.e"
#include "obj.e"
#include "pattern.e"
#include "rcbox.e"
#include "select.e"
#include "setup.e"
#include "tangram2.e"
#include "util.e"

#ifdef _TANGRAM_II

#define DO_CMD(cmd) ExecACommandFromBuffer(cmd,NULL)

static TgMenuItemInfo tangram2MenuItemInfo[] = {
   { "About Tangram-II", NULL, "Information about Tangram-II", NULL,
     CMDID_ABOUT_TANGRAM2 },
   { "Generate Markov Chain", NULL, "Generate Markov Chain", NULL,
     CMDID_TANGRAM2_GENERATE_CHAIN },
   { "Start Simulator", NULL, "Start Simulation", NULL,
     CMDID_TANGRAM2_START_SIMULATOR },
   { "Simulation Step", NULL, "Simulation Step", NULL,
     CMDID_TANGRAM2_SIMULATION_STEP },
   { "End Simulator", NULL, "End Simulation", NULL,
     CMDID_TANGRAM2_END_SIMULATOR },
   { NULL, NULL, NULL, NULL, INVALID }
};
TgMenuInfo tangram2MenuInfo={ TGMUTYPE_TEXT, tangram2MenuItemInfo,
                              CreateTangram2Menu };

/* ======================= Init & CleanUp ======================= */

void CleanUpTangram2()
{
   CleanUpTangram2ShortCut();
}

int InitTangram2()
{
   InitTangram2ShortCut();

   return TRUE;
}

/* ======================= Tangram2 Shortcuts ======================= */

static struct ShortCutRec tangram2ShortCutXlateTbl[] = {
   { '\0', 0, "AboutTangram2()", 0, CMDID_ABOUT_TANGRAM2 },
   { '\0', 0, "Tangram2GenerateChain()", 1, CMDID_TANGRAM2_GENERATE_CHAIN },
   { '\0', 0, "Tangram2StartSimulator()", 1, CMDID_TANGRAM2_START_SIMULATOR },
   { '\0', 0, "Tangram2SimulationStep()", 0, CMDID_TANGRAM2_SIMULATION_STEP },
   { '\0', 0, "Tangram2EndSimulator()", 0, CMDID_TANGRAM2_END_SIMULATOR },
   { '\0', 0, "", 0, 0 }
};

static int tangram2ShortCutCmdIdIndex[MAXTANGRAM2CMDIDS-CMDID_TANGRAM2_BASE];

void CleanUpTangram2ShortCut()
{
}

int InitTangram2ShortCut()
{
   int i=0, num_cmdids=MAXTANGRAM2CMDIDS-CMDID_TANGRAM2_BASE;

   for (i=0; i < num_cmdids; i++) {
      tangram2ShortCutCmdIdIndex[i] = INVALID;
   }
   for (i=0; *(tangram2ShortCutXlateTbl[i].name) != '\0'; i++) {
      int cmdid=tangram2ShortCutXlateTbl[i].cmdid;

      if (cmdid != INVALID) {
         int index=cmdid-CMDID_TANGRAM2_BASE;

         if (tangram2ShortCutCmdIdIndex[index] == INVALID) {
            tangram2ShortCutCmdIdIndex[index] = i;
         } else {
            fprintf(stderr, "Warning:  duplicate cmdid %1d.\n", cmdid);
         }
      }
   }
   return TRUE;
}

int ValidTangram2CmdName(buf, len, pn_num_args)
   char *buf;
   int len, *pn_num_args;
   /*
    * returns 0 if no match
    * otherwise, returns (CMDID_TANGRAM2_BASE|index) where index is the
    *       index into tangram2ShortCutXlateTbl
    */
{
   int i;

   for (i=0; *(tangram2ShortCutXlateTbl[i].name) != '\0'; i++) {
      if (strncmp(tangram2ShortCutXlateTbl[i].name, buf, len) == 0) {
         *pn_num_args = tangram2ShortCutXlateTbl[i].num_args;

         return (i+CMDID_TANGRAM2_BASE);
      }
   }
   return 0;
}

int ValidTangram2CmdId(nCmdId)
   int nCmdId;
{
   int cmd_index=nCmdId-CMDID_TANGRAM2_BASE, xlate_index=INVALID;

   if (nCmdId >= MAXTANGRAM2CMDIDS || nCmdId < CMDID_TANGRAM2_BASE) {
      return FALSE;
   }
   xlate_index = tangram2ShortCutCmdIdIndex[cmd_index];
   if (xlate_index == INVALID ||
         tangram2ShortCutXlateTbl[xlate_index].num_args != 0) {
      return FALSE;
   }
   return TRUE;
}

int DoTangram2Cmd(nCmdId, args)
   int nCmdId;
   char *args;
   /* returns INVALID if the event can be caught by other windows */
{
   switch (nCmdId) {
   case CMDID_ABOUT_TANGRAM2: AboutTangram2(); break;
   case CMDID_TANGRAM2_GENERATE_CHAIN:  Tangram2GenerateChain(args);  break;
   case CMDID_TANGRAM2_START_SIMULATOR: Tangram2StartSimulator(args); break;
   case CMDID_TANGRAM2_SIMULATION_STEP: Tangram2SimulationStep();     break;
   case CMDID_TANGRAM2_END_SIMULATOR:   Tangram2EndSimulator();       break;
   default: break;
   }
   return BAD;
}

int FetchTangram2ShortCutNumArgs(index, pn_num_args) 
   int index, *pn_num_args;
{  
   *pn_num_args = tangram2ShortCutXlateTbl[index].num_args;
   
   return TRUE;
}

int DoTangram2ShortCut(index, args)
   int index;
   char *args;
   /* return FALSE if cannot execute shortcut */
{
   if (index < 0 || index >= MAXTANGRAM2CMDIDS-CMDID_TANGRAM2_BASE) {
      return FALSE;
   }
   if (tangram2ShortCutXlateTbl[index].num_args == 0) {
      DoTangram2Cmd(tangram2ShortCutXlateTbl[index].cmdid, NULL);
   } else {
      DoTangram2Cmd(tangram2ShortCutXlateTbl[index].cmdid, args);
   }
   return TRUE;
}

/* ======================= Tangram2 Internal Commands ======================= */

void ExecStartSimulator ARGS_DECL((struct ObjRec *, char*));
void ExecSimulateStep ARGS_DECL((struct ObjRec *, char*));
void ExecEndSimulator ARGS_DECL((struct ObjRec *, char*));

static
ExecInfo gTangram2ExecInfo[] = {
   { (NLFN*)ExecStartSimulator,      "start_simulator",                   0, 0},
   { (NLFN*)ExecSimulateStep,        "simulate_step",                     0, 0},
   { (NLFN*)ExecEndSimulator,        "end_simulator",                     0, 0},
   { NULL, NULL, 0, 0 }
};

ExecInfo *Tangram2GetExecInfo(func_name)
   char *func_name;
{
   ExecInfo *pei=NULL;

   for (pei=gTangram2ExecInfo; pei->pfunc != NULL; pei++) {
      if (strcmp(pei->func_name, func_name) == 0) {
         return pei;
      }
   }
   return NULL;
}

/* --------------- Tangram2 Internal Commands Implementation --------------- */

#define MAXVARS              20
#define MAXVARNAME          250
#define MAXCONNECTTRIES       3
#define PORT               6743

#define END_OF_SIMULATION     1
#define STEP_SIMULATION       2

#define D printf ("Msg de Debug!\n");

typedef struct tagT_statevar {
  char name[MAXVARNAME];
  int  value;
} T_statevar;

typedef struct tagT_Packet {
  int code;
  int step;
  double elapsed_time;
  int varcount;
  T_statevar vars[MAXVARS];
} T_Packet;

static int simulator_socket=INVALID;
static struct ObjRec *objGC=NULL; /* Generate Chain   button */
static struct ObjRec *objSS=NULL; /* Start Simulation button */
static struct ObjRec *objSp=NULL; /* Simulation Step  button */

static
int BadAttr_Simulator(attr_name, cmd_name)
   char *attr_name, *cmd_name;
{
   char msg[MAXSTRING+1];

   sprintf(msg, "Can not find the '%s' %s '%s' command.",
         attr_name, "attribute while executing the", cmd_name);
   MsgBox(msg, TOOL_NAME, INFO_MB);
   return FALSE;
}

void ExecStartSimulator(obj_ptr, orig_cmd)
   char *orig_cmd;
   struct ObjRec *obj_ptr;
   /* start_simulator(); */
{
    char host[ 10 ];
    char *protocol = "udp";
    struct hostent     *phe;
    struct servent     *pse;
    struct protoent    *ppe;
    struct sockaddr_in  sin;
    int sin_size;
    int s, type, max_desc;
    char buffer[ sizeof( T_Packet ) ];
    fd_set soc_pool;
    struct timeval timeout;
    int one = 1;
    char *service = "6743";

    strcpy( host, "localhost" );
    bzero( (char *)&sin, sizeof( sin ) );
    sin.sin_family = AF_INET;
    if( (pse = getservbyname( service, protocol )) != NULL )
        sin.sin_port = pse->s_port;
    else
        if( (sin.sin_port = htons( (u_short)atoi( service ) )) == 0 )
        {
            fprintf( stderr, "TGIF: Can't get \"%s\" service entry\n", service );
            return;
        }
    
    if( (phe = gethostbyname( host )) != NULL )
        bcopy( phe->h_addr, (char *)&sin.sin_addr, phe->h_length );
    else
    {
        sin.sin_addr.s_addr = inet_addr( host );
#ifdef linux
        if( sin.sin_addr.s_addr == INADDR_NONE )
        {
            fprintf( stderr, "TGIF: Can't get \"%s\" host entry\n", host );
            return;
        }
#endif /* linux */
    }
        
    if( (ppe = getprotobyname( protocol )) == NULL )
    {
        fprintf( stderr, "TGIF: Can't get \"%s\" protocol entry\n", protocol );
        return;
    }
    if( strcmp( protocol, "udp" ) == 0 )
        type = SOCK_DGRAM;
    else
        type = SOCK_STREAM;
        
    if( (s = socket( PF_INET, type, ppe->p_proto )) < 0 )
    {
        fprintf( stderr, "TGIF: Can't create socket\n" );
        return;
    }
    
    if( setsockopt( s, SOL_SOCKET, SO_REUSEADDR, (char *)&one, sizeof( one ) ) < 0 )
    {
        perror( "TGIF: REUSEADDR" );
        return;
    }
    
    if( bind( s, (struct sockaddr *)&sin, sizeof( sin ) ) < 0 )
    {
        perror( "TGIF: bin" );
        return;
    }
    
    if( type == SOCK_STREAM && listen( s, 0 ) < 0 )
    {
        fprintf( stderr, "TGIF: Can't listen on %s port\n", service );
        return;
    }
    
    FD_ZERO( &soc_pool );
    FD_SET( s, &soc_pool );
    max_desc = s + 1;
    timeout.tv_sec  = 20;
    timeout.tv_usec = 0;
    if( select( max_desc, &soc_pool, NULL, NULL, &timeout ) < 0 )
    {
        perror( "TGIF: select" );
        return;
    }
    if( FD_ISSET( s, &soc_pool ) )
    {
        sin_size = sizeof( sin );
        if( recvfrom( s, buffer, sizeof( T_Packet ), 0, (struct sockaddr *)&sin, &sin_size ) < 0 )
        {
            perror( "TGIF: recv" );
            return;
        }
        
        if( connect( s, (struct sockaddr *)&sin, sin_size ) < 0 )
        {
            fprintf( stderr, "TGIF: Can't connect to %s.%s\n", host, service );
            return;
        }
    
        sleep( 1 );
        if( send( s, buffer, sizeof( T_Packet ), 0 ) < 0 )
        {
            perror( "TGIF: send" );
            return;
        }
    
    }
    else
    {
        fprintf( stderr, "TGIF: Timeout expired\n" );
        return;
    }
    
    simulator_socket = s;
    
    printf("TGIF: Cliente Conectado.\n");
}

void ExecEndSimulator(obj_ptr, orig_cmd)
   char *orig_cmd;
   struct ObjRec *obj_ptr;
   /* end_simulator(); */
{
   T_Packet pack_out;
   int bywrite;

   if (simulator_socket == INVALID) {
      perror("invalid simulator_socket");
      return;
   }
   /* monta o pacote */
   memset((char *)&pack_out, 0, sizeof(T_Packet));
   pack_out.code = END_OF_SIMULATION;
    
   /* envia o pacote ao simulador */
   bywrite = write(simulator_socket, &pack_out, sizeof(T_Packet));
   if (bywrite < 0) {
      perror("write");
      return;
   }    
   close(simulator_socket);
   printf("TGIF: Simulation finished.\n");
}

static
void monta_buffer(char *buffer, T_Packet pack)
{
   int i;
   char temp[MAXSTRING+1];

   memset(buffer, 0, sizeof(buffer));
   for (i=0; i < pack.varcount; i++) {
      strcat(buffer, pack.vars[i].name);
      sprintf(temp, "=%d@", pack.vars[i].value);
      strcat(buffer, temp);
   }

   /* printf("monta: %s\n", buffer); */
}

static
int recv_packet(int soc, T_Packet *pack_in)
{
   int  byread, status;   /* bytes read from the socket */
   char buffer[sizeof(T_Packet)];

   /* clear and receive the packet */
   memset(buffer, 0, sizeof(T_Packet));
   byread = read(soc, buffer, sizeof(T_Packet));
   if (byread > 0) {
      while (byread < sizeof(T_Packet)) {
         status = read(soc, &(buffer[byread]), sizeof(T_Packet) - byread); 
         if (status > 0) {
            byread += status;
         } else {
            byread = status;
            break;
         }
      }      
   }
   memcpy((char*)pack_in, buffer, sizeof(T_Packet));

   if (byread <= 0) {
      perror("read");
      return(-1);
   } else {
      return(1);
   }
}

static
int send_packet(int soc, T_Packet pack_out)
{
   int  bywrite, status;
   char buffer[sizeof(T_Packet)];

   memcpy(buffer, (char*)&pack_out, sizeof(T_Packet));
   bywrite = write(soc, buffer, sizeof(T_Packet)); 
   if (bywrite > 0) {
      while (bywrite < sizeof(T_Packet)) {
         status = write(soc, &(buffer[bywrite]), sizeof(T_Packet) - bywrite); 
         if (status > 0) {
            bywrite += status;
         } else {
            bywrite = status;
            break;
         }
      }
   } 
   if (bywrite <= 0) {
      perror("write");
      return(-1);
   } else {
      return(1);
   }
}


static
void desmonta_buffer(char *buffer, T_Packet *pack)
{
   int i;
   char *token;

   /* printf("desmonta: %s\n", buffer); */

   i = 0;
   token = strtok(buffer, "=");
   strcpy(pack->vars[i].name, token);
   while (token != NULL) {
      token = strtok(NULL, "@");
      pack->vars[i].value = atoi(token);
      i++;
      token = strtok(NULL, "=");
      if (token == NULL) {
         break;
      }
      strcpy(pack->vars[i].name, token);
   }

   pack->varcount = i;
}

void ExecSimulateStep(obj_ptr, orig_cmd)
   char *orig_cmd;
   struct ObjRec *obj_ptr;
   /* simulate_step(); */
{
   char buffer[MAXVARS*sizeof(T_statevar)]; 

   struct AttrRec *attr_ptr;
   struct ObjRec *attr_owner_obj=NULL;
   static char execDummyStr[MAXSTRING+1];

   struct ObjRec *owner_obj=NULL, *named_obj;
   static char obj_name[MAXSTRING+1];

   T_Packet pack_in, pack_out;

   static int transitions = 0;
   static double tempo = 0;
   int step = 1;

   if (simulator_socket == INVALID) {
      perror("invalid simulator_socket");
      return;
   }
   /* pega o valor do atributo buffer */
   strcpy(execDummyStr, "buffer=");
   attr_ptr = FindAttrWithName(obj_ptr, execDummyStr, &attr_owner_obj);
   if (attr_ptr == NULL) {BadAttr_Simulator(execDummyStr, orig_cmd); return; }
   strcpy(buffer, attr_ptr->attr_value.s);

   /* pega o valor do atributo step */
   strcpy(execDummyStr, "Step=");
   attr_ptr = FindAttrWithName(obj_ptr, execDummyStr, &attr_owner_obj);
   if (attr_ptr == NULL) { BadAttr_Simulator(execDummyStr, orig_cmd); return; }
   step = atoi(attr_ptr->attr_value.s);

   /*
    * step nao pode ser menor do que zero (se for zero entao esta com play_mode)
    */
   if (step <= 0) {
      step = 1;
   }
   /* monta o pacote */
   memset((char *)&pack_out, 0, sizeof(T_Packet));
   pack_out.code = STEP_SIMULATION;
   pack_out.step = step;
   desmonta_buffer(buffer, &pack_out);

   /* envia o pacote ao simulador */
   if (send_packet(simulator_socket, pack_out) < 0) {
      return;
   }
   /* recebe pacote de resposta do simulador */
   if (recv_packet(simulator_socket, &pack_in) < 0) {
      return;
   }
   /* desmonta o pacote */
   tempo = pack_in.elapsed_time;
   transitions = step;
   monta_buffer(buffer, pack_in);

   /* seta o valor do atributo buffer  */
   strcpy(execDummyStr, "buffer=");
   attr_ptr = FindAttrWithName(obj_ptr, execDummyStr, &attr_owner_obj);
   if (attr_ptr == NULL) { BadAttr_Simulator(execDummyStr, orig_cmd); return; }
   ReplaceAttrFirstValue(attr_owner_obj, attr_ptr, buffer);

   /* set o valor do step_time (tempo de duracao deste step) */
   strcpy(execDummyStr, "step_time=");
   attr_ptr = FindAttrWithName(obj_ptr, execDummyStr, &attr_owner_obj);
   if (attr_ptr == NULL) { BadAttr_Simulator(execDummyStr, orig_cmd); return; }
   sprintf(execDummyStr, "%.3f", tempo);
   ReplaceAttrFirstValue(attr_owner_obj, attr_ptr, execDummyStr);

   /* atualiza o valor do atributo Time do objeto "__START_SIMULATOR__" */
   strcpy(obj_name, "__START_SIMULATOR__");
   named_obj = FindObjWithName(botObj, obj_ptr, obj_name, FALSE, FALSE,
         &owner_obj, NULL);

   strcpy(execDummyStr, "Time=");
   attr_ptr = FindAttrWithName(named_obj, execDummyStr, &attr_owner_obj);
   if (attr_ptr == NULL) { BadAttr_Simulator(execDummyStr, orig_cmd); return; }
   tempo += atof(attr_ptr->attr_value.s);
   sprintf(execDummyStr, "%.3f", tempo);
   ReplaceAttrFirstValue(attr_owner_obj, attr_ptr, execDummyStr);

   /* atualiza o valor do atributo Transitions do mesmo objeto */
   strcpy(execDummyStr, "Transitions=");
   attr_ptr = FindAttrWithName(named_obj, execDummyStr, &attr_owner_obj);
   if (attr_ptr == NULL) { BadAttr_Simulator(execDummyStr, orig_cmd); return; }
   transitions += atof(attr_ptr->attr_value.s);
   sprintf(execDummyStr, "%d", transitions);
   ReplaceAttrFirstValue(attr_owner_obj, attr_ptr, execDummyStr);
}

/* ======================= Tangram2 Menu ======================= */

int RefreshTangram2Menu(menu)
   TgMenu *menu;
{
   int ok=TRUE;

   return ok;
}

TgMenu *CreateTangram2Menu(parent_menu, x, y, menu_info, status_str_xlated)
   TgMenu *parent_menu;
   int x, y;
   TgMenuInfo *menu_info;
   int status_str_xlated; /* ignored, always 0 */
{
   TgMenu *menu=TgCreateMenuFromMenuInfo(parent_menu, x, y, menu_info, FALSE);

   if (menu != NULL) {
      if (!RefreshTangram2Menu(menu)) {
         return TgDestroyMenu(menu, TRUE);
      }
   }
   return menu;
}

int Tangram2Menu(X, Y, TrackMenubar)
   int X, Y, TrackMenubar;
{
   int rc=INVALID;
   TgMenu *menu=(tangram2MenuInfo.create_proc)(NULL, X, Y, &tangram2MenuInfo,
         INVALID);

   activeMenu = MENU_TANGRAM2;
   if (menu != NULL) {
      rc = TgMenuLoop(menu);
      TgDestroyMenu(menu, TRUE);
   }
   return rc;
}

/* ========================= Tangram2 Menu Data ========================= */


static char *GenerateChainInit[] = {
  "get_current_file(\"str\");",
  "read_file_into_attr(\"| basename $(str) .obj\", \"basename\");",
  "strcpy(file_name, \"$(basename).parser\");",
  "strcpy(str, \" \");",
  "write_attr_into_file(str, $(file_name));",
  NULL
};

static char *GenerateChainExec[] = {
  "disable_undo();",
  "exec(init);",
  "exec(write_model);",
  "exec(write_global_rewards);",
  "exec(write_indep_chains);",
  "enable_undo();",
   NULL
};

static char *GenerateChainWriteModel[] = {
  "find_obj_names(obj_list,\"\",\"Declaration=*\");",
  "get_line_in_attr(total_objs,obj_list,0);",
  "for_i(i,1,$(total_objs),1,write_obj);",
  NULL
};

static char *GenerateChainErrorParts[] = {
  "message_box(NULL,\"Objects must have these four parts: Variables_and_parameters, Events, Messages and Rewards. $(str)\",\"Mode Error\", \"stop\");",
  "exec(stop);",
  NULL
};


static char *GenerateChainStop[] = {
  "enable_undo();",
  "stop();",
  NULL
};


static char *GenerateChainAppendGlobalRewards[] = {
  "append_attr_into_file(__GLOBAL_REWARDS__.global_rewards, $(file_name));",
 NULL
};


static char *GenerateChainWriteIndependentChains[] = {
  "strcpy(str, \"Independent_Chains {\");",
  "append_attr_into_file(str,$(file_name));",
  "is_attr(result, __INDEP_CHAINS__.independent_chains);",
  "if($(result) != 0, append_indep_chains, NULL);",
  "strcpy(str, \"}\");",
  "append_attr_into_file(str,$(file_name));",
  NULL
};

static char *GenerateChainAppendIndepChains[] = {
  "append_attr_into_file(__INDEP_CHAINS__.independent_chains, $(file_name));",
  NULL
};


static char *GenerateChainWriteGlobalRewards[] = {
  "strcpy(str, \"Global_Rewards {\");",
  "append_attr_into_file(str,$(file_name));",
  "is_attr(result, __GLOBAL_REWARDS__.global_rewards);",
  "if($(result) != 0, append_global_rewards, NULL);",
  "strcpy(str, \"}\");",
  "append_attr_into_file(str,$(file_name));",
  NULL
};

static char *GenerateChainWriteMaxValues[] = {
  "strcpy(max_file_name, \"$(basename).maxvalues\");",
  "strcpy(str, \"$(__SETUP__.max_values)\");",
  "write_attr_into_file(__SETUP__.max_values, $(max_file_name));",
 NULL
};


static char *GenerateChainWriteObj[] = {
  "get_line_in_attr(obj_name,obj_list,$(i));",
  "strcpy(str, \"Object_Desc $(obj_name) (\");",
  "append_attr_into_file(str,$(file_name));",
  "is_attr(result,$(obj_name).Declaration);",
  "strcpy(str,\"Missing Declaration part in object: $(obj_name).\");",
  "if($(result) == 0,error_parts,NULL);",
  "strcpy(str, \"Declaration {\");",
  "append_attr_into_file(str,$(file_name));",
  "append_attr_into_file($(obj_name).Declaration,$(file_name));",
  "strcpy(str, \"}\");",
  "append_attr_into_file(str,$(file_name));",
  "is_attr(result,$(obj_name).Initialization);",
  "strcpy(str,\"Missing Initialization part in object: $(obj_name).\");",
  "if($(result) == 0,error_parts,NULL);",
  "strcpy(str, \"Initialization {\");",
  "append_attr_into_file(str,$(file_name));",
  "append_attr_into_file($(obj_name).Initialization,$(file_name));",
  "strcpy(str, \"}\");",
  "append_attr_into_file(str,$(file_name));",
  "is_attr(result,$(obj_name).Events);",
  "strcpy(str,\"Missing Events part in object: $(obj_name).\");",
  "if($(result) == 0,error_parts,NULL);",
  "strcpy(str, \"Events {\");",
  "append_attr_into_file(str,$(file_name));",
  "append_attr_into_file($(obj_name).Events,$(file_name));",
  "strcpy(str, \"}\");",
  "append_attr_into_file(str,$(file_name));",
  "is_attr(result,$(obj_name).Messages);",
  "strcpy(str,\"Missing Messages part in object: $(obj_name).\");",
  "if($(result) == 0,error_parts,NULL);",
  "strcpy(str, \"Messages {\");",
  "append_attr_into_file(str,$(file_name));",
  "append_attr_into_file($(obj_name).Messages,$(file_name));",
  "strcpy(str, \"}\");",
  "append_attr_into_file(str,$(file_name));",
  "is_attr(result,$(obj_name).Rewards);",
  "strcpy(str,\"Missing Rewards part in object: $(obj_name).\");",
  "if($(result) == 0,error_parts,NULL);",
  "strcpy(str, \"Rewards {\");",
  "append_attr_into_file(str,$(file_name));",
  "append_attr_into_file($(obj_name).Rewards,$(file_name));",
  "strcpy(str, \"}\");",
  "append_attr_into_file(str,$(file_name));",
  "strcpy(str, \")\");",
  "append_attr_into_file(str,$(file_name));",
  NULL
};


static char *StartSimulatorSetModeB[] = {
  "strcpy(command, \"simulate_chain $(__GENERATE_CHAIN__.basename) B $(__GENERATE_CHAIN__.dbglevel) $(Event) $(Transitions) $(Time) $(Runs) >> $(__GENERATE_CHAIN__.output) & \");",
  "launch(command);",
  NULL
};

static char *StartSimulatorSetModeI[] = {
 "strcpy(Transitions, \"0\");",
 "strcpy(Time, \"0.0\");",
 "strcpy(command, \"simulate_chain $(__GENERATE_CHAIN__.basename) I $(__GENERATE_CHAIN__.dbglevel) >> $(__GENERATE_CHAIN__.output) & \");",
 "launch(command);",
 "start_simulator();",
  NULL
};

static char *StartSimulatorErrorMode[] = {
 "message_box(NULL,\"Please, Mode has to be I(Interactive) or B(Batch)!\",\"Mode Error\", \"stop\");",
  NULL
};

static char *StartSimulatorExec[] = {
 "disable_undo();",
 "show_attr(Transitions);",
 "show_attr(Time);",
 "exec(__GENERATE_CHAIN__.init);",
 "if(\"$(Mode)\"!=\"I\"&&\"$(Mode)\"!=\"B\",error_mode,NULL);"
 "strcpy(file_name, \"$(__GENERATE_CHAIN__.file_name)\");",
 "strcpy(str, \" \");",
 "write_attr_into_file(str, $(file_name));",
 "exec(__GENERATE_CHAIN__.write_model);"
 "exec(__GENERATE_CHAIN__.write_global_rewards);"
 "exec(__GENERATE_CHAIN__.write_indep_chains);",
 "strcpy(command, \"gramatica $(file_name) $(__GENERATE_CHAIN__.basename) $(__GENERATE_CHAIN__.dbglevel) -nomaxcheck > $(__GENERATE_CHAIN__.output)\");",
 "launch(command);",
 "if(\"$(Mode)\"==\"I\",set_mode_I,NULL);",
 "if(\"$(Mode)\"==\"B\",set_mode_B,NULL);",
 "enable_undo();",
 NULL
};

static char *SimulationStepWriteVarLine[] = {
 "tokenize(state_var_tok,$(state_var_line),'=');",
 "get_line_in_attr(state_var_name,state_var_tok,1);",
 "get_line_in_attr(state_var_value,state_var_tok,2);",
 "strcpy(new_line, \"$(cur_obj).$(state_var_name)=$(state_var_value)\");",
 "strcpy(buffer,\"$(buffer)$(new_line)@\");",
  NULL
};

static char *SimulationStepLoopAnimation[] = {
 "for_i(Ani_C_Step,1,$(Ani_T_Steps),1,Do_Animation);",
 "round(sleep_time,$(Time_Scale)*$(step_time)*1000);",
 "sleep(NULL,$(sleep_time));",
  NULL
};

static char *SimulationStepAnimateObj[] = {
 "get_line_in_attr(cur_obj,obj_list,$(obj_index));",
 "exec(\"$(cur_obj).Animation\");",
  NULL
};

static char *SimulationStepDoAnimation[] = {
 "find_obj_names(obj_list,\"\",\"Animation=*\");",
 "get_line_in_attr(no_objs,obj_list,0);",
 "for_i(obj_index,1,$(no_objs),1,Animate_Obj);",
  NULL
};

static char *SimulationStepCallSimulator[] = {
  "simulate_step();",
  NULL
};

static char *SimulationStepPlayOneStep[] = {
 "set_allow_interrupt(FALSE);",
 "exec(step_simulation);",
 "set_allow_interrupt(TRUE);",
 "sleep(NULL,$(Delay));",
  NULL
};

static char *SimulationStepUpdateObjects[] = {
 "tokenize(state_var_list,$(buffer),'@');",
 "get_line_in_attr(no_state_vars,state_var_list,0);",
 "for_i(var_index,1,\"$(no_state_vars)-1\",1,update_vars);",
  NULL
};

static char *SimulationStepStop[] = {
  "stop();",
  NULL
};

static char *SimulationStepStepSimulation[] = {
 "strcpy(buffer,\"\");",
 "find_obj_names(obj_list,\"\",\"State_vars=*\");",
 "get_line_in_attr(no_objs,obj_list,0);",
 "for_i(obj_index,1,$(no_objs),1,parse_obj);",
 "exec(call_simulator);",
 "exec(update_objects);",
 "redraw_drawing_area();",
 "if(\"$(Animation)\"==\"ON\",Loop_Animation,NULL);",
  NULL
};

static char *SimulationStepPlaySimulation[] = {
  "while(1,play_one_step);",
  NULL
};

static char *SimulationStepUpdateVars[] = {
 "get_line_in_attr(state_var_str,state_var_list,$(var_index));",
 "tokenize(state_var_tok,$(state_var_str),'.');",
 "get_line_in_attr(obj_name,state_var_tok,1);",
 "get_line_in_attr(state_var_str,state_var_tok,2);",
 "tokenize(state_var_tok,$(state_var_str),'=');",
 "get_line_in_attr(state_var_name,state_var_tok,1);",
 "get_line_in_attr(state_var_value,state_var_tok,2);",
 "append_line_into_attr($(obj_name).State_vars,\" $(state_var_name)=$(state_var_value)\");",
  NULL
};

static char *SimulationStepParseObj[] = {
 "get_line_in_attr(cur_obj,obj_list,$(obj_index));",
 "get_number_of_lines_in_attr(no_state_vars,$(cur_obj).State_vars);",
 "for_i(var_index,1,$(no_state_vars)-1,1,write_vars);",
 "clear_attr($(cur_obj).State_vars);",
  NULL
};

static char *SimulationStepExec[] = {
 "disable_undo();",
 "show_attr(Step);",
 "show_attr(Delay);",
 "show_attr(Animation);",
 "show_attr(Time_Scale);",
 "if(\"$(Step)\"==\"0\",play_simulation,step_simulation);",
 "enable_undo();",
 "redraw_drawing_area();",
  NULL
};

static char *SimulationStepWriteVars[] = {
 "get_line_in_attr(state_var_line,$(cur_obj).State_vars,$(var_index));",
 "trim(state_var_line);",
 "strlen(line_size, $(state_var_line));",
 "if(\"$(line_size)\">=\"3\",write_var_line,NULL);",
  NULL
};


static
void AppendLinesToAttr(attr_ptr, ppsz_lines)
   struct AttrRec *attr_ptr;
   char **ppsz_lines;
{
   char **ppsz=NULL;
   struct TextRec *text_ptr=attr_ptr->obj->detail.t;

   for (ppsz=ppsz_lines; *ppsz != NULL; ppsz++) {
      MiniLineInfo *pFirstMiniLine=NULL, *pLastMiniLine=NULL;

      CreateMiniLineFromString(*ppsz, &pFirstMiniLine, &pLastMiniLine);
      text_ptr->minilines.last->next = pFirstMiniLine;
      pFirstMiniLine->prev = text_ptr->minilines.last;
      text_ptr->minilines.last = pLastMiniLine;
      text_ptr->lines++;
   }
   RecalcTextMetrics(text_ptr, attr_ptr->obj->x, text_ptr->baseline_y);
   UpdTextBBox(attr_ptr->obj);
}

/* ======================= Tangram2 Auxiliar Functions ======================= */

struct ObjRec *createGenerateChainButton( dbglevel, maxstates, output )
char *dbglevel;
char *maxstates;
char *output;
{
   struct ObjRec *tmp_box_obj=NULL;

   CreateBoxObj(0,0,10,10,TRUE);
   tmp_box_obj = topObj;

   /* Note: AddAttrByNameAndValue() adds the new object at tmp_box_obj->fattr */
   AddAttrByNameAndValue(tmp_box_obj, "i=", "");
   AddAttrByNameAndValue(tmp_box_obj, "dbglevel=", dbglevel);
   AddAttrByNameAndValue(tmp_box_obj, "maxstates=", maxstates);
   AddAttrByNameAndValue(tmp_box_obj, "output=", output);
   AddAttrByNameAndValue(tmp_box_obj, "processID=", "");
   AddAttrByNameAndValue(tmp_box_obj, "name=", "__GENERATE_CHAIN__");
   AddAttrByNameAndValue(tmp_box_obj, "basename=", "");
   AddAttrByNameAndValue(tmp_box_obj, "total_objs=", "");
   AddAttrByNameAndValue(tmp_box_obj, "file_name=", "");
   AddAttrByNameAndValue(tmp_box_obj, "command=", "");
   AddAttrByNameAndValue(tmp_box_obj, "str=", "");
   AddAttrByNameAndValue(tmp_box_obj, "result=", "");
   AddAttrByNameAndValue(tmp_box_obj, "obj_name=", "");
   AddAttrByNameAndValue(tmp_box_obj, "max_file_name=", "");
   AddAttrByNameAndValue(tmp_box_obj, "obj_list=", "");

   AddAttrByNameAndValue(tmp_box_obj, "init=", "");
   AppendLinesToAttr(tmp_box_obj->fattr, GenerateChainInit);

   AddAttrByNameAndValue(tmp_box_obj, "exec=", "");
   AppendLinesToAttr(tmp_box_obj->fattr, GenerateChainExec);

   AddAttrByNameAndValue(tmp_box_obj, "write_model=", "");
   AppendLinesToAttr(tmp_box_obj->fattr, GenerateChainWriteModel);

   AddAttrByNameAndValue(tmp_box_obj, "write_obj=", "");
   AppendLinesToAttr(tmp_box_obj->fattr, GenerateChainWriteObj);

   AddAttrByNameAndValue(tmp_box_obj, "stop=", "");
   AppendLinesToAttr(tmp_box_obj->fattr, GenerateChainStop);

   AddAttrByNameAndValue(tmp_box_obj, "error_parts=", "");
   AppendLinesToAttr(tmp_box_obj->fattr, GenerateChainErrorParts);

   AddAttrByNameAndValue(tmp_box_obj, "write_max_values=", "");
   AppendLinesToAttr(tmp_box_obj->fattr, GenerateChainWriteMaxValues);

   AddAttrByNameAndValue(tmp_box_obj, "write_global_rewards=", "");
   AppendLinesToAttr(tmp_box_obj->fattr, GenerateChainWriteGlobalRewards);

   AddAttrByNameAndValue(tmp_box_obj, "append_global_rewards=", "");
   AppendLinesToAttr(tmp_box_obj->fattr, GenerateChainAppendGlobalRewards);

   AddAttrByNameAndValue(tmp_box_obj, "write_indep_chains=", "");
   AppendLinesToAttr(tmp_box_obj->fattr, GenerateChainWriteIndependentChains);

   AddAttrByNameAndValue(tmp_box_obj, "append_indep_chains=", "");
   AppendLinesToAttr(tmp_box_obj->fattr, GenerateChainAppendIndepChains);

   AddNewSelObj(tmp_box_obj);

   return( tmp_box_obj );
}


struct ObjRec *createStartSimulatorButton( runs, event, trans, time, mode )
char *runs;
char *event;
char *trans;
char *time;
char *mode;
{
   struct ObjRec *tmp_box_obj = NULL;
   int y = 5;

   CreateRCBoxObj(4,y,200,y+40,TRUE);
   tmp_box_obj = topObj;

   tmp_box_obj->color = 9;
   tmp_box_obj->detail.rcb->width = 2;
   y += 5;

   AddAttrByNameAndValue(tmp_box_obj, "Runs=", runs);
   AddAttrByNameAndValue(tmp_box_obj, "Event=", event);
   AddAttrByNameAndValue(tmp_box_obj, "Transitions=", trans);
   tmp_box_obj->fattr->shown = TRUE;
   MoveObj(tmp_box_obj->fattr->obj, 20, y - tmp_box_obj->fattr->obj->obbox.lty);
   y += 14;

   AddAttrByNameAndValue(tmp_box_obj, "Time=", time);
   tmp_box_obj->fattr->shown = TRUE;
   MoveObj(tmp_box_obj->fattr->obj, 20, y - tmp_box_obj->fattr->obj->obbox.lty);
   y += 14;

   AddAttrByNameAndValue(tmp_box_obj, "Mode=", mode);
   AddAttrByNameAndValue(tmp_box_obj, "file_name=", "");
   AddAttrByNameAndValue(tmp_box_obj, "command=", "");
   AddAttrByNameAndValue(tmp_box_obj, "simulating=", "");
   AddAttrByNameAndValue(tmp_box_obj, "str=", "");
   AddAttrByNameAndValue(tmp_box_obj, "name=", "__START_SIMULATOR__");

   AddAttrByNameAndValue(tmp_box_obj, "set_mode_B=", "");
   AppendLinesToAttr(tmp_box_obj->fattr, StartSimulatorSetModeB);

   AddAttrByNameAndValue(tmp_box_obj, "set_mode_I=", "");
   AppendLinesToAttr(tmp_box_obj->fattr, StartSimulatorSetModeI);

   AddAttrByNameAndValue(tmp_box_obj, "error_mode=", "");
   AppendLinesToAttr(tmp_box_obj->fattr, StartSimulatorErrorMode);

   AddAttrByNameAndValue(tmp_box_obj, "exec=", "");
   AppendLinesToAttr(tmp_box_obj->fattr, StartSimulatorExec);

   AdjObjBBox(tmp_box_obj);

   AddNewSelObj(tmp_box_obj);
   RedrawDrawWindow(botObj);

   return( tmp_box_obj );
}

struct ObjRec *createSimulationStepButton( time, anim, delay, step )
char *time;
char *anim;
char *delay;
char *step;
{
   struct ObjRec *tmp_box_obj = NULL;

   CreateBoxObj(0,0,10,10,TRUE);
   tmp_box_obj = topObj;

   /* Note: AddAttrByNameAndValue() adds the new object at tmp_box_obj->fattr */
   AddAttrByNameAndValue(tmp_box_obj, "Time_Scale=", time);
   AddAttrByNameAndValue(tmp_box_obj, "Animation=", anim);
   AddAttrByNameAndValue(tmp_box_obj, "Delay=", delay);
   AddAttrByNameAndValue(tmp_box_obj, "Step=", step);
   AddAttrByNameAndValue(tmp_box_obj, "state_var_str=", "1");
   AddAttrByNameAndValue(tmp_box_obj, "line_size=", "7");
   AddAttrByNameAndValue(tmp_box_obj, "obj_list=", "");
   AddAttrByNameAndValue(tmp_box_obj, "sleep_time=", "2049");
   AddAttrByNameAndValue(tmp_box_obj, "step_time=", "2049");
   AddAttrByNameAndValue(tmp_box_obj, "name=", "__SIMULATION_STEP__");
   AddAttrByNameAndValue(tmp_box_obj, "cur_obj=", "");
   AddAttrByNameAndValue(tmp_box_obj, "obj_index=", "");
   AddAttrByNameAndValue(tmp_box_obj, "no_objs=", "");
   AddAttrByNameAndValue(tmp_box_obj, "Ani_T_Steps=", "10");
   AddAttrByNameAndValue(tmp_box_obj, "Ani_C_Step=", "11");
   AddAttrByNameAndValue(tmp_box_obj, "step_index=", "");
   AddAttrByNameAndValue(tmp_box_obj, "no_state_vars=", "3");
   AddAttrByNameAndValue(tmp_box_obj, "state_var_value=", "8");
   AddAttrByNameAndValue(tmp_box_obj, "state_var_name=", "8");
   AddAttrByNameAndValue(tmp_box_obj, "obj_name=", "");
   AddAttrByNameAndValue(tmp_box_obj, "new_line=", "");
   AddAttrByNameAndValue(tmp_box_obj, "buffer=", "");
   AddAttrByNameAndValue(tmp_box_obj, "state_var_line=", "");
   AddAttrByNameAndValue(tmp_box_obj, "state_var_list=", "");
   AddAttrByNameAndValue(tmp_box_obj, "state_var_tok=", "");
   AddAttrByNameAndValue(tmp_box_obj, "var_index=", "");

   AddAttrByNameAndValue(tmp_box_obj, "write_var_line=", "");
   AppendLinesToAttr(tmp_box_obj->fattr, SimulationStepWriteVarLine);

   AddAttrByNameAndValue(tmp_box_obj, "Loop_Animation=", "");
   AppendLinesToAttr(tmp_box_obj->fattr, SimulationStepLoopAnimation);

   AddAttrByNameAndValue(tmp_box_obj, "Animate_Obj=", "");
   AppendLinesToAttr(tmp_box_obj->fattr, SimulationStepAnimateObj);

   AddAttrByNameAndValue(tmp_box_obj, "Do_Animation=", "");
   AppendLinesToAttr(tmp_box_obj->fattr, SimulationStepDoAnimation);

   AddAttrByNameAndValue(tmp_box_obj, "call_simulator=", "");
   AppendLinesToAttr(tmp_box_obj->fattr, SimulationStepCallSimulator);

   AddAttrByNameAndValue(tmp_box_obj, "play_one_step=", "");
   AppendLinesToAttr(tmp_box_obj->fattr, SimulationStepPlayOneStep);

   AddAttrByNameAndValue(tmp_box_obj, "update_objects=", "");
   AppendLinesToAttr(tmp_box_obj->fattr, SimulationStepUpdateObjects);

   AddAttrByNameAndValue(tmp_box_obj, "stop=", "");
   AppendLinesToAttr(tmp_box_obj->fattr, SimulationStepStop);

   AddAttrByNameAndValue(tmp_box_obj, "step_simulation=", "");
   AppendLinesToAttr(tmp_box_obj->fattr, SimulationStepStepSimulation);

   AddAttrByNameAndValue(tmp_box_obj, "play_simulation=", "");
   AppendLinesToAttr(tmp_box_obj->fattr, SimulationStepPlaySimulation);

   AddAttrByNameAndValue(tmp_box_obj, "update_vars=", "");
   AppendLinesToAttr(tmp_box_obj->fattr, SimulationStepUpdateVars);

   AddAttrByNameAndValue(tmp_box_obj, "parse_obj=", "");
   AppendLinesToAttr(tmp_box_obj->fattr, SimulationStepParseObj);

   AddAttrByNameAndValue(tmp_box_obj, "exec=", "");
   AppendLinesToAttr(tmp_box_obj->fattr, SimulationStepExec);

   AddAttrByNameAndValue(tmp_box_obj, "write_vars=", "");
   AppendLinesToAttr(tmp_box_obj->fattr, SimulationStepWriteVars);

   AddNewSelObj(tmp_box_obj);

   return( tmp_box_obj );
}

/******************************************************************************/
/* This function extracts the args that are comma separated in 'args' and     */
/* stores them in 'dst'.                                                      */
/* return 0 if failed, no_args if success.                                    */
/* arguments:                                                                 */
/*    args: comma separated list of tokens                                    */
/*    dst : char vector pointer where the tokens will be stored               */
/*    no_args: number of tokens in 'args'                                     */
/*    max_arg_length: max token's size                                        */
/******************************************************************************/
int parseArgs( args, dst, no_args, max_arg_length )
char *args;
char *dst;
int no_args;
int max_arg_length;
{
    int i, j;
    char *ptr;

    bzero( dst, no_args * max_arg_length );    
    ptr = args;
    /* Remove leading " and spaces */
    while( *ptr == '\"' || *ptr == ' ' )
        ptr++;
        
    for( i = 0; i < no_args; i++ )
    {
        j = 0;
        while( *ptr != '\0' && *ptr != ';' && *ptr != '\"' &&
               j < max_arg_length )
            *(dst + max_arg_length * i + j++) = *ptr++;
            
        if( *ptr == ';' || *ptr == '\"' )
            ptr++;
        else
            if( *ptr != '\0' )
                break;
    }
    
    return( i == no_args );
}
/* ======================= Tangram2 Menu Functions ======================= */

void AboutTangram2()
{
   sprintf(gszMsgBox, "%s\n\n%s %s.",
         "Welcome to Tangram-II!",
         "Brought to you by the Tangram-II team at the",
         "Universidade Federal do Rio de Janeiro");
   MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
}

/******************************************************************************/
/* This function creates a invisible button, executes this button and removes */
/* it. This function also needs to extract three arguments that are colapsed  */
/* in arg. This is done because tgif supports only one argument functions.    */
/******************************************************************************/
void Tangram2GenerateChain( args )
char *args; /* semi-colon separated list (arg1;arg2;arg3) */
{
   int saved_fill=objFill, saved_pen=penPat;
   struct ObjRec *tmp_box_obj=NULL;
   struct AttrRec *exec_attr=NULL;
   char arg[3][1024];

   /* Extracts parameters from args */
   if( !parseArgs( args, arg, 3, 1024 ) )
   {
        fprintf( stderr, "TGIF: Usage Tangram2GenerateChain( \"arg1;arg2;arg3\" )\n" );
        return;
   }

   MakeQuiescent(); /* select nothing, edit nothing */

   objFill = penPat = NONEPAT;

   tmp_box_obj = createGenerateChainButton( arg[0], arg[1], arg[2] );

   objFill = saved_fill;
   penPat = saved_pen;

   exec_attr = FindAttrWithName(topSel->obj, "exec=", NULL);
   if (exec_attr != NULL) {
      DoExecLoop(topSel->obj, exec_attr);
   }
   MakeQuiescent();
   AddNewSelObj(tmp_box_obj);
   PrepareToRecord(CMD_DELETE, topSel, botSel, numObjSelected);
   UnlinkObj(topSel->obj);
   FreeObj(topSel->obj);
   RemoveAllSel();
   RecordCmd(CMD_DELETE, NULL, NULL, NULL, 0);

}

void Tangram2StartSimulator( args )
char *args;
{
   int saved_fill=objFill, saved_pen=penPat;
   struct AttrRec *exec_attr=NULL;
   char arg[12][1024]; /* first 3 for generate_chain */

   /* Extracts parameters from args */
   if( !parseArgs( args, arg, 12, 1024 ) )
   {
        return;
   }

   MakeQuiescent(); /* select nothing, edit nothing */
   objFill = penPat = NONEPAT;

   objGC = createGenerateChainButton( arg[0], arg[1], arg[2] );
   objSp = createSimulationStepButton( arg[8], arg[9], arg[10], arg[11] );

   objFill = saved_fill;
   penPat = saved_pen;
   objSS = createStartSimulatorButton( arg[3], arg[4], arg[5], arg[6], arg[7] );

   exec_attr = FindAttrWithName(objSS, "exec=", NULL);
   if (exec_attr != NULL) {
      DoExecLoop(objSS, exec_attr);
   }
}

void Tangram2SimulationStep()
{
   struct AttrRec *exec_attr=NULL;
   
   exec_attr = FindAttrWithName(objSp, "exec=", NULL);
   if (exec_attr != NULL) {
      DoExecLoop(objSp, exec_attr);
   }
}

void Tangram2EndSimulator()
{
   ExecEndSimulator( NULL, NULL );

   /* Delete all created buttons */
   MakeQuiescent();
   AddNewSelObj(objGC);
   AddNewSelObj(objSS);
   AddNewSelObj(objSp);
   DelAllSelObj();

   RedrawDrawWindow(botObj);
}

#endif /* _TANGRAM_II */
